/**
 * @file	main.c
 * @brief	N クイーン問題の解の個数を求めるコマンド
 * @note	ビット演算未使用のシングル・スレッド版。
 * @author	SHINDOH Yasutaka
 * @date	2006/09/12
 */

/************************************************************
 * インクルード・ファイルの読み込み
 ************************************************************/

#include <stddef.h>		/* NULL, size_t */
#include <stdio.h>		/* stdout, stderr, fprintf(), fputc() */
#include <string.h>		/* strcmp(), memset() */
#include <assert.h>		/* assert() */

/************************************************************
 * C99 対応
 * - C99 環境の場合、USE_EXPANDABLE_ARRAYS を定義することで、
 *   可変長配列を用い、処理対象の盤の幅の上限を無くせる。
 * - C99 環境の場合、inline 関数を使用している箇所があるが、
 *   USE_FUNCTIONAL_MACROS を定義することで、
 *   関数形式マクロに置き換えることができる。
 ************************************************************/

#if	!defined(__STDC_VERSION__) || (__STDC_VERSION__ < 199901L)
/**
 * @brief	関数形式マクロを使用
 */
#define	USE_FUNCTIONAL_MACROS
#ifdef	USE_EXPANDABLE_ARRAYS
#error	"可変長配列を使用するには、C99 環境が必要です。"
#endif	/* USE_EXPANDABLE_ARRAYS */
#endif	/* 非 C99 環境 */

/************************************************************
 * 型関連・関数形式マクロ
 ************************************************************/

/**
 * @brief	論理型
 * @note	通常は char を、USE_INTEGER_BOOL 定義時は size_t を使用する。
 * @note	高速性を目指す場合、型の選択は環境によって変更すること。
 */
#ifdef	USE_INTEGER_BOOL
typedef int BOOL;
#else	/* USE_INTEGER_BOOL */
typedef char BOOL;
#endif	/* USE_INTEGER_BOOL */

/**
 * @brief	論理型の false
 */
#define	FALSE	((BOOL)0)

/**
 * @brief	論理型の true
 */
#define	TRUE	((BOOL)~(BOOL)0)

/************************************************************
 * 定数定義
 ************************************************************/

/**
 * @brief	盤のデフォルト幅
 */
#define	DEFAULT_BOARD_WIDTH	((size_t)8)

#ifndef	USE_EXPANDABLE_ARRAYS
/**
 * @brief	盤の最大幅
 * @note	USE_EXPANDABLE_ARRAYS 未定義時に使用。
 */
#define	MAX_BOARD_WIDTH	((size_t)20)
#endif	/* USE_EXPANDABLE_ARRAYS */

/************************************************************
 * 非メイン関数定義
 ************************************************************/

/**
 * @brief	盤の状態を出力
 * @param	file [out] 出力先
 * @param	put_columns [in] クイーンの配置位置
 * @param	width [in] 盤の幅
 * @param	count [in] 盤の表示の前に付ける正の数
 * @note	引数 count が負の数の場合、数字のつかない盤の表示になる。
 * @note	不適切な引数を適用した場合の動作は不定。
 */
static void
Print(FILE* const file,
	  const size_t* const put_columns,
	  const size_t width,
	  const size_t count)
{
	size_t column, c;
	size_t l = 0;
	assert(file);
	assert(put_columns);
	if (0 < count) {
		fprintf(file, ">>%u\n", (unsigned int)count);
	}
	else {
		fprintf(file, ">>>>\n");
	}
	while (l < width) {
		column = put_columns[l++];
		for (c = 0; c < width; ++c) {
			if (column != c) {
				fputc('-', file);
			}
			else {
				fputc('Q', file);
			}
		}
		fputc('\n', file);
	}
}

/**
 * @brief	クイーンを置ける可能性を確認
 * @param	put_columns [out] クイーンの効き筋の情報 (列)
 * @param	put_lh_diagonal [out] クイーンの効き筋の情報 (左斜め上)
 * @param	put_rh_diagonal [out] クイーンの効き筋の情報 (右斜め上)
 * @param	line [in] 行
 * @param	column [in] 列
 * @retval	非 FALSE: 置ける
 * @retval	FALSE: 置けない
 * @note	不適切な引数を適用した場合の動作は不定。
 */
#ifdef	USE_FUNCTIONAL_MACROS
#define	PutQueenOk(pc,plh,prh,l,c) \
	(!(*((pc)+(c)) | *((plh)+(l)-(c)) | *((prh)+(l)+(c))))
#else	/* USE_FUNCTIONAL_MACROS */
inline BOOL
PutQueenOk(const BOOL* const put_columns,
		   const BOOL* const put_lh_diagonal,
		   const BOOL* const put_rh_diagonal,
		   const size_t line,
		   const size_t column)
{
	assert(put_columns);
	assert(put_lh_diagonal);
	assert(put_rh_diagonal);
	return !(*(put_columns+column) | *(put_lh_diagonal+line-column) | *(put_rh_diagonal+line+column));
}
#endif	/* USE_FUNCTIONAL_MACROS */

/**
 * @brief	クイーンを置く
 * @param	put_columns [out] クイーンの効き筋の情報 (列)
 * @param	put_lh_diagonal [out] クイーンの効き筋の情報 (左斜め上)
 * @param	put_rh_diagonal [out] クイーンの効き筋の情報 (右斜め上)
 * @param	line [in] 行
 * @param	column [in] 列
 * @note	不適切な引数を適用した場合の動作は不定。
 */
#ifdef	USE_FUNCTIONAL_MACROS
#define	PutQueen(pc,plh,prh,l,c) \
	{ *((prh)+(l)+(c)) = *((plh)+(l)-(c)) = *((pc)+(c)) = TRUE; }
#else	/* USE_FUNCTIONAL_MACROS */
inline void
PutQueen(BOOL* const put_columns,
		 BOOL* const put_lh_diagonal,
		 BOOL* const put_rh_diagonal,
		 const size_t line,
		 const size_t column)
{
	assert(put_columns);
	assert(put_lh_diagonal);
	assert(put_rh_diagonal);
	assert(!*(put_columns+column));
	assert(!*(put_lh_diagonal+line-column));
	assert(!*(put_rh_diagonal+line+column));
	*(put_rh_diagonal+line+column) = *(put_lh_diagonal+line-column) = *(put_columns+column) = TRUE;
}
#endif	/* USE_FUNCTIONAL_MACROS */

/**
 * @brief	クイーンをどかす
 * @param	put_columns [out] クイーンの効き筋の情報 (列)
 * @param	put_lh_diagonal [out] クイーンの効き筋の情報 (左斜め上)
 * @param	put_rh_diagonal [out] クイーンの効き筋の情報 (右斜め上)
 * @param	line [in] 行
 * @param	column [in] 列
 * @note	不適切な引数を適用した場合の動作は不定。
 */
#ifdef	USE_FUNCTIONAL_MACROS
#define	RemoveQueen(pc,plh,prh,l,c) \
	{ *((prh)+(l)+(c)) = *((plh)+(l)-(c)) = *((pc)+(c)) = FALSE; }
#else	/* USE_FUNCTIONAL_MACROS */
inline void
RemoveQueen(BOOL* const put_columns,
			BOOL* const put_lh_diagonal,
			BOOL* const put_rh_diagonal,
			const size_t line,
			const size_t column)
{
	assert(put_columns);
	assert(put_lh_diagonal);
	assert(put_rh_diagonal);
	assert(*(put_columns+column));
	assert(*(put_lh_diagonal+line-column));
	assert(*(put_rh_diagonal+line+column));
#if	defined(__i386__) || defined(__x86_64__)	/* x86, x64 */
	*(put_rh_diagonal+line+column) ^= *(put_rh_diagonal+line+column);
	*(put_lh_diagonal+line-column) ^= *(put_lh_diagonal+line-column);
	*(put_columns+column) ^= *(put_columns+column);
#else	/* x86, x64 */
	*(put_rh_diagonal+line+column) = *(put_lh_diagonal+line-column) = *(put_columns+column) = FALSE;
#endif	/* x86, x64 */
}
#endif	/* USE_FUNCTIONAL_MACROS */

/**
 * @brief	N クイーンの解を求める (盤の状態の出力あり)
 * @param	file [out] 出力先
 * @param	width [in] 盤の幅
 * @note	C99 でない場合、引数 width は MAX_BOARD_WIDTH 以下にすること。
 */
size_t
GetSolutions(FILE* const file,
			 const size_t width)
{
	size_t count = 0;
	assert(file);
#ifndef	USE_EXPANDABLE_ARRAYS
	assert(width<=MAX_BOARD_WIDTH);
#endif	/* !USE_EXPANDABLE_ARRAYS */
	if (width > 0) {
		int line = 0;
#ifdef	USE_EXPANDABLE_ARRAYS
		size_t columns[width];
		BOOL put_array[width*5-2];
#else	/* USE_EXPANDABLE_ARRAYS */
		size_t columns[MAX_BOARD_WIDTH];
		BOOL put_array[MAX_BOARD_WIDTH*5-2];
#endif	/* USE_EXPANDABLE_ARRAYS */
		BOOL* const put_c = put_array;
		BOOL* const put_lh = put_array + width * 2 - 1;
		BOOL* const put_rh = put_array + width * 3 - 1;
		memset((void*)columns, 0, sizeof(size_t)*width);
		memset((void*)put_array, 0, sizeof(BOOL)*width*5-2);
		while (TRUE) {
			if (columns[line] < width) {
				if (!PutQueenOk(put_c, put_lh, put_rh, (size_t)line, columns[line])) {
					columns[line]++;
				}
				else {
					PutQueen(put_c, put_lh, put_rh, (size_t)line, columns[line]);
					if (++line >= (int)width) {
						++count;
						Print(file, columns, width, count);
						--line;
						RemoveQueen(put_c, put_lh, put_rh, (size_t)line, columns[line]);
						columns[line]++;
					}
				}
			}
			else {
				columns[line] = 0;
				if (--line < 0) break;
				RemoveQueen(put_c, put_lh, put_rh, (size_t)line, columns[line]);
				columns[line]++;
			}
		}
	}
	return count;
}

/**
 * @brief	N クイーンの解を求める (表示なし)
 * @param	width [in] 盤の幅
 * @note	C99 でない場合、引数 width は MAX_BOARD_WIDTH 以下にすること。
 */
size_t
GetNumberOfSolutions(const size_t width)
{
	size_t count = 0;
#ifndef	USE_EXPANDABLE_ARRAYS
	assert(width<=MAX_BOARD_WIDTH);
#endif	/* !USE_EXPANDABLE_ARRAYS */
	if (width > 1) {
		const size_t half_width = width / 2;
		size_t line, i;
#ifdef	USE_EXPANDABLE_ARRAYS
		size_t columns[width];
		BOOL put_array[width*5-2];
#else	/* USE_EXPANDABLE_ARRAYS */
		size_t columns[MAX_BOARD_WIDTH];
		BOOL put_array[MAX_BOARD_WIDTH*5-2];
#endif	/* USE_EXPANDABLE_ARRAYS */
		BOOL* const put_c = put_array;
		BOOL* const put_lh = put_array + width * 2 - 1;
		BOOL* const put_rh = put_array + width * 3 - 1;
		memset((void*)columns, 0, sizeof(size_t)*width);
		memset((void*)put_array, 0, sizeof(BOOL)*width*5-2);
		for (i = 0; i < half_width; ++i) {
			PutQueen(put_c, put_lh, put_rh, 0, i);
			line = 1;
			while (TRUE) {
				if (columns[line] < width) {
					if (!PutQueenOk(put_c, put_lh, put_rh, line, columns[line])) {
						columns[line]++;
					}
					else {
						PutQueen(put_c, put_lh, put_rh, line, columns[line]);
						if (++line >= width) {
							count += 2;
							--line;
							RemoveQueen(put_c, put_lh, put_rh, line, columns[line]);
							columns[line]++;
						}
					}
				}
				else {
					columns[line] = 0;
					if (--line < 1) break;
					RemoveQueen(put_c, put_lh, put_rh, line, columns[line]);
					columns[line]++;
				}
			}
			RemoveQueen(put_c, put_lh, put_rh, 0, i);
		}
		if (width % 2 > 0) {
			const size_t loop = half_width - 1;
			PutQueen(put_c, put_lh, put_rh, 0, half_width);
			for (i = 0; i < loop; ++i) {
				PutQueen(put_c, put_lh, put_rh, 1, i);
				line = 2;
				while (TRUE) {
					if (columns[line] < width) {
						if (!PutQueenOk(put_c, put_lh, put_rh, line, columns[line])) {
							columns[line]++;
						}
						else {
							PutQueen(put_c, put_lh, put_rh, line, columns[line]);
							if (++line >= width) {
								count += 2;
								--line;
								RemoveQueen(put_c, put_lh, put_rh, line, columns[line]);
								columns[line]++;
							}
						}
					}
					else {
						columns[line] = 0;
						if (--line < 2) break;
						RemoveQueen(put_c, put_lh, put_rh, line, columns[line]);
						columns[line]++;
					}
				}
				RemoveQueen(put_c, put_lh, put_rh, 1, i);
			}
			RemoveQueen(put_c, put_lh, put_rh, 0, half_width);
		}
	}
	else if (width == 1) {
		++count;
	}
	return count;
}

/************************************************************
 * メイン関数定義
 ************************************************************/

/**
 * @brief	メイン関数
 * @note	コマンドとして呼ぶ際にオプションとして
 *			-p や -print を与えると解の盤を表示します。
 *			数字を与えると、その幅・高さを持つ盤で探索します。
 *			-h や -help を与えると使用方法を表示します。
 */
int
main(int arg_count,
	 char** arg_strings)
{
	size_t width = DEFAULT_BOARD_WIDTH;
	size_t i, number;
	const char* arg_string = NULL;
	BOOL print = FALSE;

	/************************************************************/
	/* オプション処理 *******************************************/
	/************************************************************/

	for (i = 1; i < (size_t)arg_count; ++i) {
		arg_string = arg_strings[i];
		if ('0' < *arg_string && *arg_string <= '9') {
			/* board width */
			number = 0;
			do {
				if (*arg_string < '0' || '9' < *arg_string) goto error;
				number = number * 10 + (size_t)(*arg_string - '0');
			} while (*++arg_string != '\0');
#ifndef	USE_EXPANDABLE_ARRAYS
			if (MAX_BOARD_WIDTH < number) goto error;
#endif	/* USE_EXPANDABLE_ARRAYS */
			width = number;
			continue;
		}
		if (*arg_string == '-') {
			/* other option */
			++arg_string;
			if (*arg_string == '-') {
				/* long option */
				++arg_string;
				if (!strcmp(arg_string,"print")) {
					print = TRUE;
				}
				else if (!strcmp(arg_string,"help")) {
					goto help;
				}
				else {
					goto error;
				}
			}
			else {
				/* short option */
				do {
					switch (*arg_string) {
					case 'p':
						print = TRUE;
						break;
					case 'h':
						goto help;
					default:
						goto error;
					}
				} while (*++arg_string != '\0');
			}
			continue;
		}
		goto error;
	}

	/************************************************************/
	/* 主処理 ***************************************************/
	/************************************************************/

	if (print) {
		number = GetSolutions(stdout, width);
		fprintf(stdout, ">>RESULT\n");
	}
	else {
		number = GetNumberOfSolutions(width);
	}
	fprintf(stdout, "解の数 (%d クイーン): %u\n",
			(int)width, (unsigned int)number);
	return 0;

	/************************************************************/
	/* 使用方法表示 *********************************************/
	/************************************************************/

error:	/* unknown option */
	fprintf(stderr, "\"%s\" は不正なオプションです。\n", arg_strings[i]);
help:	/* help */
#ifdef	USE_EXPANDABLE_ARRAYS
	fprintf(stderr,
			"Usage: %s [-ph] [number]\n"
			"\tnumber\t\tボードの幅 (デフォルト値: %d)\n"
			"\t-p, --print\t解のボードをプリント\n"
			"\t-h, --help\tこのヘルプをプリント\n",
			*arg_strings, (int)DEFAULT_BOARD_WIDTH);
#else	/* USE_EXPANDABLE_ARRAYS */
	fprintf(stderr,
			"Usage: %s [-ph] [number]\n"
			"\tnumber\t\tボードの幅 (デフォルト: %d, 最大: %d)\n"
			"\t-p, --print\t解のボードをプリント\n"
			"\t-h, --help\tこのヘルプをプリント\n",
			*arg_strings, (int)DEFAULT_BOARD_WIDTH, (int)MAX_BOARD_WIDTH);
#endif	/* USE_EXPANDABLE_ARRAYS */
	return 1;
}
